<?php

namespace Drupal\searchstax\Plugin\SolrConnector;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\search_api\Backend\BackendInterface;
use Drupal\search_api\ServerInterface;
use Drupal\search_api_solr\Controller\SolrConfigSetController;
use Drupal\search_api_solr\Plugin\SolrConnector\StandardSolrConnector;
use Drupal\search_api_solr\SearchApiSolrException;
use Solarium\Core\Client\Endpoint;
use Solarium\Core\Client\Response;
use Symfony\Component\DependencyInjection\ContainerInterface;

// cspell:ignore configset solrcore ulog

/**
 * Provides a plugin for connecting to a SearchStax Solr server with token auth.
 *
 * @SolrConnector(
 *   id = "searchstax",
 *   label = @Translation("SearchStax Cloud with Token Auth"),
 *   description = @Translation("Index items protected by token authentication for SearchStax."),
 * )
 */
class SearchStaxConnector extends StandardSolrConnector {

  /**
   * Cached version strings, keyed by server ID.
   *
   * @var array<string, string>
   */
  protected static array $versionStrings = [];

  /**
   * Cached Solr config files, keyed by server ID.
   *
   * @var array<string, array<string, string>>
   */
  protected static array $cachedFiles = [];

  /**
   * The server to which this connector plugin is linked.
   */
  protected ServerInterface $server;

  /**
   * The Solr config set controller.
   */
  protected SolrConfigSetController $solrConfigSetController;

  /**
   * The entity type manager.
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * Constructs a new class instance.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\search_api_solr\Controller\SolrConfigSetController $solr_config_set_controller
   *   The Solr config set controller.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    array $plugin_definition,
    SolrConfigSetController $solr_config_set_controller,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->solrConfigSetController = $solr_config_set_controller;
    $this->entityTypeManager = $entity_type_manager;

    // Try to retrieve the server entity from the call stack.
    $this->setServerFromBacktrace();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ): self {
    if ($container->has('search_api_solr.configset_controller')) {
      $solr_config_set_controller = $container->get('search_api_solr.configset_controller');
    }
    else {
      $solr_config_set_controller = new SolrConfigSetController($container->get('extension.list.module'));
    }
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $solr_config_set_controller,
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'scheme' => 'https',
      'host' => '',
      'port' => 443,
      'context' => '',
      'core' => '',
      'update_endpoint' => '',
      'update_token' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['update_endpoint'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Endpoint'),
      '#description' => $this->t('Just copy &amp; paste the “Update endpoint” value of the SearchStax app / index as shown in your SearchStax account.'),
      '#default_value' => $this->configuration['update_endpoint'] ?? '',
      '#required' => TRUE,
    ];

    $form['update_token'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Read &amp; write token key'),
      '#description' => $this->t('Just copy &amp; paste the “Read & Write token key” value of the SearchStax app / index as shown in your SearchStax account.'),
      '#default_value' => $this->configuration['update_token'] ?? '',
      '#required' => TRUE,
    ];

    $form += parent::buildConfigurationForm($form, $form_state);

    $form['scheme'] = [
      '#type' => 'value',
      '#value' => 'https',
    ];

    $form['host'] = [
      '#type' => 'value',
      '#value' => '',
    ];

    $form['port'] = [
      '#type' => 'value',
      '#value' => '443',
    ];

    $form['path'] = [
      '#type' => 'value',
      '#value' => '/',
    ];

    $form['core'] = [
      '#type' => 'value',
      '#value' => '',
    ];

    $form['context'] = [
      '#type' => 'value',
      '#value' => '',
    ];

    $form['advanced']['jmx'] = [
      '#type' => 'value',
      '#value' => FALSE,
    ];

    $form['advanced']['solr_install_dir'] = [
      '#type' => 'value',
      '#value' => '',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $values = $form_state->getValues();
    $values['update_endpoint'] = trim($values['update_endpoint']);
    $form_state->setValue('update_endpoint', $values['update_endpoint']);
    $values['update_token'] = trim($values['update_token']);
    $form_state->setValue('update_token', $values['update_token']);

    if (preg_match('@https://([^/:]+)/([^/]+)/([^/]+)/(update|select)$@', $values['update_endpoint'], $matches)) {
      $form_state->setValue('host', $matches[1]);
      $form_state->setValue('context', $matches[2]);
      $form_state->setValue('core', $matches[3]);
    }
    else {
      $form_state->setErrorByName('update_endpoint', $this->t('Invalid endpoint format'));
    }

    if (empty($values['update_token'])) {
      $form_state->setErrorByName('update_token', $this->t('Invalid token format'));
    }

    // Check that the necessary method for setting the authorization token
    // exists.
    if (!method_exists(Endpoint::class, 'setAuthorizationToken')) {
      $form_state->setErrorByName('update_token', $this->t('The version of the Solarium library installed on your site does not support token authentication. Upgrade to a version newer than @version to use this Solr connector.', ['@version' => '6.2.7']));
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function connect(): void {
    if (!$this->solr) {
      parent::connect();
      $this->solr->getEndpoint('search_api_solr')
        ->setAuthorizationToken('Token', $this->configuration['update_token']);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function pingServer() {
    return $this->pingCore();
  }

  /**
   * {@inheritdoc}
   */
  public function getServerInfo($reset = FALSE): array {
    // @todo the required APIs to auto-detect the Solr version are blocked by
    //   SearchStax. So we have to hardcode a better minimal version number
    //   unless an API becomes available. That's better than falling back to
    //   6.0.0 which will result in the usage of deprecated field types in the
    //   config-set.
    return ['lucene' => ['solr-spec-version' => '8.11.1']];
  }

  /**
   * {@inheritdoc}
   */
  public function getSchemaVersionString($reset = FALSE): string {
    $server_id = $this->getServer()->id();
    if (!isset(static::$versionStrings[$server_id])) {
      static::$versionStrings[$server_id] = 'drupal-0.0.0-solr-8.x';

      $this->connect();
      $query = $this->solr->createApi([
        'handler' => $this->configuration['core'] . '/schema',
      ]);

      if ($response = $this->execute($query)->getResponse()) {
        $body = json_decode($response->getBody(), TRUE);
        if (isset($body['schema']['name'])) {
          static::$versionStrings[$server_id] = $body['schema']['name'];
        }
      }
    }

    return static::$versionStrings[$server_id];
  }

  /**
   * {@inheritdoc}
   */
  public function coreRestGet($path, ?Endpoint $endpoint = NULL): array {
    return [
      'fieldTypes' => [
        [
          'name' => 'Information about fieldTypes is not provided by SearchStax',
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function reloadCore(): bool {
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function getLuke(): array {
    return [
      'fields' => [],
      'index' => ['numDocs' => -1],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getStatsSummary(): array {
    return [
      '@pending_docs' => 0,
      '@index_size' => 0,
      '@schema_version' => $this->getSchemaVersionString(),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFile($file = NULL) {
    $server = $this->getServer();
    $server_id = $server->id();
    if (!isset(static::$cachedFiles[$server_id])) {
      $this->solrConfigSetController->setServer($server);
      $files = $this->solrConfigSetController->getConfigFiles();
      foreach ($files as $name => $content) {
        $content = preg_replace('/"drupal-\d+\.\d+\.\d+[^"]+"/m', '"' . $this->getSchemaVersionString() . '"', $content);
        $files[$name] = [
          'body' => $content,
          'size' => strlen($content),
        ];
      }
      ksort($files);
      static::$cachedFiles[$server_id] = $files;
    }

    if (!$file) {
      return new Response(json_encode([
        'files' => static::$cachedFiles[$server_id],
      ]));
    }
    if (empty(static::$cachedFiles[$server_id][$file])) {
      throw new SearchApiSolrException('File not found');
    }
    return new Response(static::$cachedFiles[$server_id][$file]['body']);
  }

  /**
   * {@inheritdoc}
   */
  public function alterConfigFiles(array &$files, string $lucene_match_version, string $server_id = ''): void {
    parent::alterConfigFiles($files, $lucene_match_version, $server_id);

    if (strpos($files['solrconfig.xml'], 'numVersionBuckets') === FALSE) {
      $files['solrconfig.xml'] = str_replace('</updateLog>', '<int name="numVersionBuckets">${solr.ulog.numVersionBuckets:65536}</int>' . "\n</updateLog>", $files['solrconfig.xml']);
    }
    $files['solrconfig.xml'] = str_replace('{solr.autoCommit.MaxTime:15000}', '{solr.autoCommit.MaxTime:600000}', $files['solrconfig.xml']);
    $files['solrconfig.xml'] = str_replace('{solr.autoSoftCommit.MaxTime:5000}', '{solr.autoSoftCommit.maxTime:300000}', $files['solrconfig.xml']);

    // Leverage the implicit Solr request handlers with default settings for
    // Solr Cloud.
    // @see https://lucene.apache.org/solr/guide/8_0/implicit-requesthandlers.html
    $files['solrconfig_extra.xml'] = preg_replace("@<requestHandler\s+name=\"/replication\".*?</requestHandler>@s", '', $files['solrconfig_extra.xml']);
    $files['solrconfig_extra.xml'] = preg_replace("@<requestHandler\s+name=\"/get\".*?</requestHandler>@s", '', $files['solrconfig_extra.xml']);

    // Set the StatsCache.
    // @see https://lucene.apache.org/solr/guide/8_0/distributed-requests.html#configuring-statscache-distributed-idf
    if (!empty($this->configuration['stats_cache'])) {
      $files['solrconfig_extra.xml'] .= '<statsCache class="' . $this->configuration['stats_cache'] . '" />' . "\n";
    }

    // solrcore.properties won’t work in SolrCloud mode (it is not read from
    // ZooKeeper). Therefore, we go for a more specific fallback to keep the
    // possibility to set the property as parameter of the virtual machine.
    // @see https://lucene.apache.org/solr/guide/8_6/configuring-solrconfig-xml.html
    $files['solrconfig.xml'] = preg_replace('/solr.luceneMatchVersion:LUCENE_\d+/', 'solr.luceneMatchVersion:' . $this->getLuceneMatchVersion(), $files['solrconfig.xml']);
    unset($files['solrcore.properties']);
  }

  /**
   * Retrieves the server to which this connector plugin is linked.
   *
   * @return \Drupal\search_api\ServerInterface
   *   The server.
   *
   * @throws \Drupal\search_api_solr\SearchApiSolrException
   *   Thrown if the server could not be determined.
   */
  protected function getServer(): ServerInterface {
    // First try to retrieve the server from the call stack.
    if (!isset($this->server)) {
      $this->setServerFromBacktrace();
    }
    // If that doesn't work, the only remaining approach is to load all
    // SearchStax servers and match against their connector configs.
    if (!isset($this->server)) {
      try {
        $server_storage = $this->entityTypeManager->getStorage('search_api_server');
        $server_ids = $server_storage->getQuery()
          ->condition('backend', 'search_api_solr')
          ->condition('backend_config.connector', 'searchstax')
          ->execute();
        if ($server_ids) {
          /** @var \Drupal\search_api\ServerInterface $server */
          foreach ($server_storage->loadMultiple($server_ids) as $server) {
            if (($server->getBackendConfig()['connector_config'] ?? []) === $this->configuration) {
              $this->server = $server;
              break;
            }
          }
        }
      }
      catch (\Exception $ignored) {
      }
    }
    // If that didn't work, either, we can only throw an exception to at least
    // prevent a fatal error.
    if (!isset($this->server)) {
      throw new SearchApiSolrException('Could not determine server for connector plugin.');
    }
    return $this->server;
  }

  /**
   * Attempts to extract the plugin's server from the backtrace.
   */
  protected function setServerFromBacktrace(): void {
    $options = DEBUG_BACKTRACE_PROVIDE_OBJECT | DEBUG_BACKTRACE_IGNORE_ARGS;
    $limit = 8;
    $trace = debug_backtrace($options, $limit);
    foreach ($trace as $call) {
      $object = $call['object'] ?? NULL;
      if ($object instanceof ServerInterface) {
        $this->server = $object;
        break;
      }
      if ($object instanceof BackendInterface) {
        $this->server = $object->getServer();
        break;
      }
    }
  }

}
